//
//  Typesetter.m
//  WriteRoom
//
//  Created by Jesse Grosjean on 7/11/15.
//  Copyright (c) 2015 Hog Bay Software. All rights reserved.
//

#import "Typesetter.h"
#import "BulletAttachmentCell.h"
#import "TextContainer.h"
#import "LayoutManager.h"
#import "ItemRenderer.h"
#import "TextStorage.h"
#import "StyleSheet.h"
#import "Logging.h"

@implementation Typesetter {
	ItemRenderer *_currentItemRenderer;
}

- (instancetype)init {
	return [super init];
}

- (void)dealloc {
	DDLogVerbose(@"dealloc: Typesetter");
}

- (ItemRenderer *)currentItemRenderer {
	if (!_currentItemRenderer) {
		TextStorage *textStorage = (id) self.layoutManager.textStorage;
        if ([textStorage isKindOfClass:[TextStorage class]]) {
            NSRange range = self.paragraphCharacterRange;
            if (range.location == textStorage.length) {
                JSValue *lastDisplayed = [textStorage.editor invokeMethod:@"getLastDisplayedItem" withArguments:@[]];
				if (!lastDisplayed.isUndefined) {
					_currentItemRenderer = [[ItemRenderer alloc] initWithItem:lastDisplayed editor:textStorage.editor];
				}
            } else {
                _currentItemRenderer = [textStorage attribute:ItemRendererAttributeName atIndex:self.paragraphCharacterRange.location effectiveRange:NULL];
            }
        }
	}
	return _currentItemRenderer;
}

- (void)beginParagraph {
	_currentItemRenderer = nil;
	[super beginParagraph];
}

- (void)endParagraph {
	_currentItemRenderer = nil;
	[super endParagraph];
}

- (NSDictionary *)attributesForExtraLineFragment {
    TextStorage *textStorage = (id) self.layoutManager.textStorage;
    if ([textStorage isKindOfClass:[TextStorage class]]) {
        return textStorage.editorComputedTextStyle;
    } else {
        return nil;
    }
}

- (NSUInteger)paragraphGlyphRangeLocation {
    // Do it this way because for some reason self.paragraphGlyphRange is getting flagged as private api in Mac App Store.
    return [[self valueForKey:@"paragraphGlyphRange"] rangeValue].location;
}

- (void)getLineFragmentRect:(NSRectPointer)lineFragmentRect usedRect:(NSRectPointer)lineFragmentUsedRect remainingRect:(NSRectPointer)remainingRect forStartingGlyphAtIndex:(NSUInteger)startingGlyphIndex proposedRect:(NSRect)proposedRect lineSpacing:(CGFloat)lineSpacing paragraphSpacingBefore:(CGFloat)paragraphSpacingBefore paragraphSpacingAfter:(CGFloat)paragraphSpacingAfter {
	
	TextContainer *textContainer = (id) self.currentTextContainer;
	ItemRenderer *currentItemRenderer = self.currentItemRenderer;
	if (currentItemRenderer) {
		textContainer.currentItemDepth = currentItemRenderer.depth;
	} else {
		textContainer.currentItemDepth = textContainer.hoistedDepth + 1;
	}

	[super getLineFragmentRect:lineFragmentRect usedRect:lineFragmentUsedRect remainingRect:remainingRect forStartingGlyphAtIndex:startingGlyphIndex proposedRect:proposedRect lineSpacing:lineSpacing paragraphSpacingBefore:paragraphSpacingBefore paragraphSpacingAfter:paragraphSpacingAfter];
    
    NSUInteger paragraphGlyphRangeLocation = [self paragraphGlyphRangeLocation];
	if (currentItemRenderer && paragraphGlyphRangeLocation != startingGlyphIndex) {
		if ([currentItemRenderer.attributes[@"data-type"] isEqualToString:@"task"]) {
            NSPoint start = [self.layoutManager locationForGlyphAtIndex:paragraphGlyphRangeLocation];
            NSPoint end = [self.layoutManager locationForGlyphAtIndex:paragraphGlyphRangeLocation + 2];
            end.x -= start.x;
			lineFragmentRect->origin.x += end.x;
			lineFragmentRect->size.width -= end.x;
		}
	}
}

- (void)willSetLineFragmentRect:(NSRectPointer)lineRect forGlyphRange:(NSRange)glyphRange usedRect:(NSRectPointer)usedRect baselineOffset:(CGFloat *)baselineOffset {
    ItemRenderer *currentItemRenderer = self.currentItemRenderer;
    NSParagraphStyle *paragraphStyle = self.currentParagraphStyle;
    NSFont *font = [self.attributedString attribute:NSFontAttributeName atIndex:self.paragraphCharacterRange.location effectiveRange:NULL];
    CGFloat unusedHeight = lineRect->size.height - usedRect->size.height;
    CGFloat beforeSpacing = 0;
    CGFloat afterSpacing = 0;
    
    if (unusedHeight && glyphRange.location != 0 && glyphRange.location == [self paragraphGlyphRangeLocation]) {
        beforeSpacing = paragraphStyle.paragraphSpacingBefore;
        unusedHeight -= beforeSpacing;
    }
    
    if (unusedHeight) {
        afterSpacing = paragraphStyle.paragraphSpacing;
    }
    
    lineRect->size.width = usedRect->size.width + (usedRect->origin.x - lineRect->origin.x);
    
    if (currentItemRenderer) {
        lineRect->size.width += [(id)layoutManager paragraphSeparatorAdvanceForItemRenderer:currentItemRenderer];
    }
    
    *baselineOffset = ((lineRect->size.height + beforeSpacing - afterSpacing) / 2.0) + (font.xHeight / 2.0);
}

#pragma mark -
#pragma mark Glyph Storage

- (void)deleteGlyphsInRange:(NSRange)glyphRange {
	[super deleteGlyphsInRange:glyphRange];
}

- (void)substituteGlyphsInRange:(NSRange)glyphRange withGlyphs:(NSGlyph *)glyphs {
	[super substituteGlyphsInRange:glyphRange withGlyphs:glyphs];
}

- (void)insertGlyph:(NSGlyph)glyph atGlyphIndex:(NSUInteger)glyphIndex characterIndex:(NSUInteger)characterIndex {
	[super insertGlyph:glyph atGlyphIndex:glyphIndex characterIndex:characterIndex];
}

- (void)setAttachmentSize:(NSSize)attachmentSize forGlyphRange:(NSRange)glyphRange {
    [super setAttachmentSize:attachmentSize forGlyphRange:glyphRange];
}

- (void)setLocation:(NSPoint)location withAdvancements:(const CGFloat *)advancements forStartOfGlyphRange:(NSRange)glyphRange {
	[super setLocation:location withAdvancements:advancements forStartOfGlyphRange:glyphRange];
}

@end
